"""Package for things which are loaded from outside Galaxy and which we can be
sure will not conflict with the `galaxy` namespace, which may be provided by
other packages (e.g. galaxy_ops in the toolshed).
"""
